#include "puzzle.h"

#include <iostream>
#include <vector>
#include <utility>
#include <algorithm>

using namespace std;

const int ALL_DIRS = 4;

const int LEFT = 0;
const int RIGHT = 1;
const int UP = 2;
const int DOWN = 3;

const int DESTROY_BRIDGE = 0;

const pair<int, int> NOT_AVAILABLE = {-1, -1};

const int dirs[ALL_DIRS][2] = {{0, -1}, {0, +1}, {-1, 0}, {+1, 0}};

struct Island {
    int power;
    int bridgesToBuild;

    int x, y;

    Island(int _power, int _x, int _y): power(_power), bridgesToBuild(_power), x(_x), y(_y) {}

    bool operator<(const Island &comp) const {
        return power > comp.power;
    }
};

struct Neighbor {
    int islandIndex;
    int power;

    int direction;
    vector<int> possibleBridges;

    Neighbor(int _islandIndex, int _power, int _direction): islandIndex(_islandIndex), power(_power), direction(_direction), possibleBridges({}) {}

    bool operator<(const Neighbor &comp) const {
        return power > comp.power;
    }
};

int subtestsCounter;
double testTimeLimit;

bool is_path_free(pair<int, int> &startIsland, pair<int, int> &endIsland, int dir, vector<vector<int>> &puzzle) {
    pair<int, int> currentPosition = {startIsland.first + dirs[dir][0], startIsland.second + dirs[dir][1]};

    while (currentPosition != endIsland) {
        if (puzzle[currentPosition.first][currentPosition.second] < 0) {
            return false;
        }
     
        currentPosition.first += dirs[dir][0];
        currentPosition.second += dirs[dir][1];
    }

    return true;
}

void build_bridge(pair<int, int> &startIsland, pair<int, int> &endIsland, int dir, int bridgeType, vector<vector<int>> &puzzle) {
    pair<int, int> currentPosition = {startIsland.first + dirs[dir][0], startIsland.second + dirs[dir][1]};

    while (currentPosition != endIsland) {
        puzzle[currentPosition.first][currentPosition.second] = bridgeType;
    
        currentPosition.first += dirs[dir][0];
        currentPosition.second += dirs[dir][1];
    }
}

bool solve_puzzle(int currentIsland, vector<Island> &islands, int currentNeighbor, vector<vector<Neighbor>> neighbors, vector<vector<int>> &puzzle) {
    if (timePassed() > testTimeLimit) {
        return false;
    }

    if (currentIsland == (int) islands.size()) {
        return true;
    }

    if (islands[currentIsland].bridgesToBuild == 0) {
        return solve_puzzle(currentIsland + 1, islands, 0, neighbors, puzzle);
    }

    if (currentNeighbor == (int) neighbors[currentIsland].size()) {
        return false;
    }

    pair<int, int> currentPosition = {islands[currentIsland].x, islands[currentIsland].y};

    for (int i = currentNeighbor; i < (int) neighbors[currentIsland].size(); ++i) {
        Island &nextIsland = islands[neighbors[currentIsland][i].islandIndex];

        bool isOneWayBridgePossible = nextIsland.bridgesToBuild >= 1;
        bool isTwoWayBridgePossible = nextIsland.bridgesToBuild >= 2;

        if (!isOneWayBridgePossible) {
            continue;
        }

        pair<int, int> nextPosition = {nextIsland.x, nextIsland.y};
        int currentDirection = neighbors[currentIsland][i].direction;

        if (!is_path_free(currentPosition, nextPosition, currentDirection, puzzle)) {
            continue;
        }

        for (int newBridge : neighbors[currentIsland][i].possibleBridges) {
            if (newBridge == HOR_DOUBLE or newBridge == VERT_DOUBLE) {
                if (!isTwoWayBridgePossible) {
                    continue;
                }

                build_bridge(currentPosition, nextPosition, currentDirection, newBridge, puzzle);
                islands[currentIsland].bridgesToBuild -= 2;
                nextIsland.bridgesToBuild -= 2;

                if (solve_puzzle(currentIsland, islands, currentNeighbor + 1, neighbors, puzzle)) {
                    return true;
                }

                build_bridge(currentPosition, nextPosition, currentDirection, DESTROY_BRIDGE, puzzle);
                islands[currentIsland].bridgesToBuild += 2;
                nextIsland.bridgesToBuild += 2;
            } else {
                build_bridge(currentPosition, nextPosition, currentDirection, newBridge, puzzle);
                islands[currentIsland].bridgesToBuild -= 1;
                nextIsland.bridgesToBuild -= 1;

                if (solve_puzzle(currentIsland, islands, currentNeighbor + 1, neighbors, puzzle)) {
                    return true;
                }

                build_bridge(currentPosition, nextPosition, currentDirection, DESTROY_BRIDGE, puzzle);
                islands[currentIsland].bridgesToBuild += 1;
                nextIsland.bridgesToBuild += 1;
            }
        }
    }

    return false;
}

void init(int _numSubtests, double _timeLimit) {
    subtestsCounter = _numSubtests;
    testTimeLimit = _timeLimit;
}

bool solve(vector<vector<int>>& puzzle) {
    int rows = (int) puzzle.size();
    int cols = (int) puzzle[0].size();

    vector<vector<vector<pair<int, int>>>> closestIsland(ALL_DIRS, vector<vector<pair<int, int>>>(rows, vector<pair<int, int>>(cols)));

    for (int row = 0; row < rows; ++row) {
        pair<int, int> currentClosest = NOT_AVAILABLE;

        for (int col = 0; col < cols; ++col) {
            if (puzzle[row][col] == 0) {
                continue;
            }

            closestIsland[LEFT][row][col] = currentClosest;
            currentClosest = {row, col};
        }

        currentClosest = NOT_AVAILABLE;

        for (int col = cols - 1; col >= 0; --col) {
            if (puzzle[row][col] == 0) {
                continue;
            }

            closestIsland[RIGHT][row][col] = currentClosest;
            currentClosest = {row, col};
        }
    }

    for (int col = 0; col < cols; ++col) {
        pair<int, int> currentClosest = NOT_AVAILABLE;

        for (int row = 0; row < rows; ++row) {
            if (puzzle[row][col] == 0) {
                continue;
            }

            closestIsland[UP][row][col] = currentClosest;
            currentClosest = {row, col};
        }

        currentClosest = NOT_AVAILABLE;

        for (int row = rows - 1; row >= 0; --row) {
            if (puzzle[row][col] == 0) {
                continue;
            }

            closestIsland[DOWN][row][col] = currentClosest;
            currentClosest = {row, col};
        }
    }

    vector<Island> islands;
    

    for (int row = 0; row < rows; ++row) {
        for (int col = 0; col < cols; ++col) {
            if (puzzle[row][col] != 0) {
                islands.push_back(Island(puzzle[row][col], row, col));
            }
        }
    }

    int allIslands = (int) islands.size();

    sort(islands.begin(), islands.end());

    vector<vector<int>> coordinatesToIsland(rows, vector<int>(cols));

    for (int i = 0; i < allIslands; ++i) {
        coordinatesToIsland[islands[i].x][islands[i].y] = i;
    }

    vector<vector<Neighbor>> neighbors(allIslands); 

    for (int i = 0; i < allIslands; ++i) {
        bool doubleBridgeAllowed = islands[i].power > 1;
        
        int x = islands[i].x;
        int y = islands[i].y;

        vector<Neighbor> currentNeighbors;

        for (int dir = 0; dir < ALL_DIRS; ++dir) {
            if (closestIsland[dir][x][y] == NOT_AVAILABLE) {
                continue;
            }

            int neighborIndex = coordinatesToIsland[closestIsland[dir][x][y].first][closestIsland[dir][x][y].second];

            Neighbor newNeighbor = Neighbor(neighborIndex, islands[neighborIndex].power, dir);

            if (doubleBridgeAllowed) {
                newNeighbor.possibleBridges.push_back(dir < 2 ? HOR_DOUBLE : VERT_DOUBLE);
            }

            newNeighbor.possibleBridges.push_back(dir < 2 ? HOR_SINGLE : VERT_SINGLE);
        
            currentNeighbors.push_back(newNeighbor);
        }

        sort(currentNeighbors.begin(), currentNeighbors.end());
    
        neighbors[i] = currentNeighbors;
    }

    return solve_puzzle(0, islands, 0, neighbors, puzzle);
}